package gs.smolban.db.doobie

import cats.MonadError
import cats.data.EitherT
import cats.syntax.all.*
import doobie.*
import doobie.implicits.*
import gs.smolban.db.DbError
import gs.smolban.db.TagDb
import gs.smolban.db.doobie.DoobieTypes.*
import gs.smolban.model.CreatedAt
import gs.smolban.model.Tag

final class DoobieTagDb() extends TagDb[ConnectionIO]:

  import DoobieTagDb.Sql
  import gs.smolban.db.DbConstants

  /** @inheritdoc
    */
  override def createTag(
    value: Tag.Value,
    createdAt: CreatedAt
  ): EitherT[ConnectionIO, DbError, Tag] =
    EitherT(
      Sql.createTag(value, createdAt).option.attemptSql.flatMap {
        case Left(ex) =>
          if DbConstants.UniqueConstraintFailed.contains(ex.getErrorCode()) then
            Left(
              DbError.TagAlreadyExists(value)
            ).pure[ConnectionIO]
          else MonadError[ConnectionIO, Throwable].raiseError(ex)
        case Right(None) =>
          Left(DbError.SilentInsertFailure("doobie-tag-db")).pure[ConnectionIO]
        case Right(Some(newId)) =>
          Right(Tag(newId, value, createdAt)).pure[ConnectionIO]
      }
    )

  /** @inheritdoc
    */
  override def readTag(id: Tag.Id): ConnectionIO[Option[Tag]] =
    Sql.readTag(id).option

  /** @inheritdoc
    */
  override def deleteTag(id: Tag.Id): ConnectionIO[Boolean] =
    for rows <- Sql.deleteTag(id).run
    yield rows > 0

object DoobieTagDb:

  private object Sql:

    def createTag(
      tv: Tag.Value,
      createdAt: CreatedAt
    ): Query0[Tag.Id] = sql"""
    INSERT INTO tags (tag_value, created_at)
    VALUES ($tv, $createdAt)
    RETURNING id
    """.query[Tag.Id]

    def readTag(tagId: Tag.Id): Query0[Tag] = sql"""
    SELECT id, value, created_at
    FROM tags
    WHERE id = $tagId
    """.query[Tag]

    def deleteTag(tagId: Tag.Id): Update0 = sql"""
    DELETE FROM tags
    WHERE id = $tagId
    """.update

  end Sql

end DoobieTagDb
