package gs.smolban.db.doobie

import DoobieTypes.ErrorMessages
import doobie.*
import gs.slug.v0.Slug
import gs.smolban.model.CreatedAt
import gs.smolban.model.Group
import gs.smolban.model.Tag
import gs.smolban.model.Ticket
import gs.uuid.v0.UUID
import gs.uuid.v0.UUIDFormat
import java.time.Instant

trait DoobieTypes:

  implicit val ticketNumberGet: Get[Ticket.Number] =
    Get[Long].tmap(Ticket.Number(_))

  implicit val ticketNumberPut: Put[Ticket.Number] =
    Put[Long].tcontramap(_.toLong())

  implicit val ticketStatusGet: Get[Ticket.Status] =
    Get[String].temap(dbValue =>
      Ticket.Status.parse(dbValue) match
        case None         => Left(ErrorMessages.invalidTicketStatus(dbValue))
        case Some(status) => Right(status)
    )

  implicit val ticketStatusPut: Put[Ticket.Status] =
    Put[String].tcontramap(_.name)

  implicit val groupIdGet: Get[Group.Id] =
    Get[Array[Byte]].tmap(bytes => Group.Id(UUID(UUIDFormat.fromBytes(bytes))))

  implicit val groupIdPut: Put[Group.Id] =
    Put[Array[Byte]].tcontramap(id => UUIDFormat.toBytes(id.toUUID().toUUID()))

  implicit val createdAtGet: Get[CreatedAt] =
    Get[Long].tmap(CreatedAt.fromMilliseconds)

  implicit val createdAtPut: Put[CreatedAt] =
    Put[Long].tcontramap(_.toMilliseconds())

  implicit val instantGet: Get[Instant] =
    Get[Long].tmap(Instant.ofEpochMilli)

  implicit val instantPut: Put[Instant] =
    Put[Long].tcontramap(_.toEpochMilli())

  implicit val slugGet: Get[Slug] =
    Get[String].temap { slug =>
      Slug.validate(slug) match
        case None    => Left(ErrorMessages.invalidSlug(slug))
        case Some(s) => Right(s)
    }

  implicit val slugPut: Put[Slug] =
    Put[String].tcontramap(_.str())

  implicit val tagIdGet: Get[Tag.Id] =
    Get[Long].tmap(Tag.Id(_))

  implicit val tagIdPut: Put[Tag.Id] =
    Put[Long].tcontramap(_.toLong())

  implicit val tagValueGet: Get[Tag.Value] =
    Get[String].temap { value =>
      Tag.Value.validate(value) match
        case None    => Left(ErrorMessages.invalidTagValue(value))
        case Some(s) => Right(s)
    }

  implicit val tagValuePut: Put[Tag.Value] =
    Put[String].tcontramap(_.toString())

object DoobieTypes extends DoobieTypes:

  object ErrorMessages:

    def invalidTagValue(candidate: String): String =
      s"'$candidate' is not a valid tag value."

    def invalidTicketStatus(candidate: String): String =
      s"'$candidate' is not a valid ticket status."

    def invalidSlug(candidate: String): String =
      s"'$candidate' is not a valid Group slug."

  end ErrorMessages

end DoobieTypes
